<?php
/**
 * Page Output
 * @package     Page
 * @copyright   Copyright (C) 2011 MicroEngine Inc.
 */
class Page_Output extends Page
{
    /** キャッシュディレクトリ */
    const CACHE_DIR = '/cache';

    /**
     * @var Page_Dao
     */
    private $dao;
    /**
     * @var Metag_Dao
     */
    private $metag_dao;
    /**
     * @var Template_Logic
     */
    private $template_logic;
    /**
     * @var Template_Parser
     */
    private $parser;
    /**
     * Document
     * @var Metag_Template
     */
    private $document;
    /**
     * Page内のmetagのデータ
     * @var array
     */
    private $data;
    /**
     * 出力ファイル
     * @var string
     */
    private $output_file;
    /**
     * モジュール用出力ファイル
     * @var array
     */
    private $module_output_list;
    /**
     * 参照ページデータ
     * @var array
     */
    private $ref_page_data;
    /**
     * 参照テンプレートデータ
     * @var array
     */
    private $ref_template_repeat;
    /**
     * @var Template_Form;
     */
    private $template_form;

    /**
     * コンストラクタ
     */
    public function __construct()
    {
        $this->dao = new Page_Dao();
        $this->metag_dao = new Metag_Dao();
        $this->template_logic = new Template_Logic();
        $this->parser = new Template_Parser();
    }

    /**
     * ページ書き出し
     * @param string $page_id
     */
    public function output($page_id)
    {
        // data初期化
        $this->data = array();

        // ページ情報取得
        $this->_id = $page_id;
        if (!$this->dao->set_page($this)) {
            return false;
        }

        if ($this->page_type !== Page_Logic::PAGE_TYPE_TEMPLATE) {
            return false;
        }
        if ($this->public_flag == 0) {
            // 非公開の場合はファイルを削除する
            $file_path = SITE_ROOT . $this->page_path;
            if (file_exists($file_path)) {
                unlink($file_path);
            }
            return true;
        }

        // テンプレート情報取得
        // template_idが同じ場合は再利用する
        if ($this->template_form == null || $this->template_form->_id !== $this->template_id) {
            $this->template_form = $this->template_logic->get_template_form_by_page($this);

            // ページ情報からテンプレートを作成
            $this->template_logic->get_template_file($this->template_form);
            // parse処理
            $this->parser->output_method = $this->template_form->output_method;
            $this->document = $this->parser->parse($this->template_form->source_code);
        }

        // ページデータ取得
        $this->set_data($this->metag_dao->get_page_data($this->_id));
        // ページ基本情報セット
        $this->set_basic_info();

        $this->output_file = '';
        if ($this->template_form->output_method === Template_Logic::OUTPUT_DYNAMIC) {
            // 先頭スクリプト追加
            $file_path = SITE_PATH . $this->page_path;
            // site.phpの絶対パス
            $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
            // index.phpへの相対パスを取得
            $relative_path = ME_Util::get_relative_path($file_path, $me_path);
            $this->output_file = "<?php include_once('" . $relative_path . "') ?>";
        }

        // 参照データを取得する
        $this->ref_page_data = array();
        foreach ($this->document->ref_page_list as $ref_page_id) {
            $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
            // idを添字にした配列に変換
            $data = $this->convert_assoc_by_id($rows);
            $this->ref_page_data[$ref_page_id] = $data;
        }
        foreach ($this->document->ref_template_repeat as $repeat) {
            $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
            $this->data[$repeat->id] = $rows;
        }

        $this->module_output_list = array();
        $this->composition($this->document);

        if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
            $this->output_file = mb_convert_encoding($this->output_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
        }

        // モジュール書き出し
        foreach ($this->module_output_list as $module_name => $module_file) {
            // ファイル名のチェック
            if (!preg_match('/^[\w-.]+$/', $module_name)) {
                // a-zA-Z-9_-. 以外の文字があれば書き出さない
                continue;
            }
            if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
                $module_file = mb_convert_encoding($module_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
            }
            // ファイル書き出し
            Me_Util::write_file(DATA_ROOT . '/module/' . $module_name, $module_file);
        }

        // メインテンプレートと指定しているサブテンプレートを取得
        $sub_template_list = $this->template_logic->select_sub_template_list($this->template_form->_id);
        if (count($sub_template_list) == 0) {
            // サブテンプレートを持たない場合
            // 公開側ファイル書き出し
            Me_Util::write_file(SITE_ROOT . $this->page_path, $this->output_file);
        } else {
            // サブテンプレートを持つ、メインテンプレートの場合
            $this->output_cache($sub_template_list);
        }
    }

    /**
     * プレビュー
     * @param Page_Form $form
     */
    public function preview(Page_Form $form)
    {
        // ページ情報取得
        $this->_id = $form->_id;
        $this->template_id = $form->template_id;
        $this->page_type = $form->page_type;
        $this->page_path = $form->page_path;
        $this->title = $form->title;
        $this->keywords = $form->keywords;
        $this->description = $form->description;
        $this->create_date = $form->create_date;
        $this->update_date = $form->update_date;

        // テンプレート情報取得
        $this->template_form = $form->template_form;
        // ソースコード再取得
        $this->template_logic->get_template_file($this->template_form);
        $this->template_form->source_code = $this->template_logic->add_base_tag(
            $this->template_form->source_code, $this->template_form->template_path);
        // parse処理
        $this->parser->output_method = Template_Logic::OUTPUT_STATIC;
        $this->document = $this->parser->parse($this->template_form->source_code);

        // ページデータセット。空のリピート領域を削除する。
        $this->set_data($this->delete_empty_repeat($form));
        // ページ基本情報セット
        $this->set_basic_info();

        $this->output_file = '';

        // 参照データを取得する
        $this->ref_page_data = array();
        foreach ($this->document->ref_page_list as $ref_page_id) {
            $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
            // idを添字にした配列に変換
            $data = $this->convert_assoc_by_id($rows);
            $this->ref_page_data[$ref_page_id] = $data;
        }
        foreach ($this->document->ref_template_repeat as $repeat) {
            $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
            $this->data[$repeat->id] = $rows;
        }

        $this->composition($this->document);

        if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
            $this->output_file = mb_convert_encoding($this->output_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
        }

        // ファイル出力
        h::no_escape($this->output_file);
    }

    /**
     * キャッシュファイル書き出し
     * @param array $sub_template_list 
     */
    private function output_cache($sub_template_list)
    {
        $cache_list = array();
        // メインテンプレートのキャッシュファイル書き出し
        Me_Util::write_file(DATA_ROOT . self::CACHE_DIR . $this->page_path, $this->output_file);
        foreach ($sub_template_list as $sub_template) {
            $sub_template_form = $this->template_logic->select($sub_template['_id'], 'Template_Form');

            // ページ情報からテンプレートを作成
            $this->template_logic->get_template_file($sub_template_form);
            // parse処理
            $this->parser->output_method = $sub_template_form->output_method;
            $sub_template_document = $this->parser->parse($sub_template_form->source_code);

            $this->output_file = '';
            if ($sub_template_form->output_method === Template_Logic::OUTPUT_DYNAMIC) {
                // 先頭スクリプト追加
                $file_path = SITE_PATH . $this->page_path;
                // site.phpの絶対パス
                $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
                // index.phpへの相対パスを取得
                $relative_path = ME_Util::get_relative_path($file_path, $me_path);
                $this->output_file = "<?php include_once('" . $relative_path . "') ?>";
            }

            // 参照データを取得する
            foreach ($sub_template_document->ref_page_list as $ref_page_id) {
                // メインテンプレートで取得していない場合だけ、データを取得する
                if (!isset($this->ref_page_data[$ref_page_id])) {
                    $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
                    // idを添字にした配列に変換
                    $data = $this->convert_assoc_by_id($rows);
                    $this->ref_page_data[$ref_page_id] = $data;
                }
            }
            foreach ($sub_template_document->ref_template_repeat as $repeat) {
                // limit, offsetなど、メインテンプレートと指定が違う可能性があるので、取得し直す。
                $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
                $this->data[$repeat->id] = $rows;
            }

            $this->module_output_list = array();
            $this->composition($sub_template_document);

            if ($sub_template_form->char_code !== null && $sub_template_form->char_code !== Template_Logic::CHAR_CODE) {
                $this->output_file = mb_convert_encoding($this->output_file, $sub_template_form->char_code, Template_Logic::CHAR_CODE);
            }
            // サブテンプレートのキャッシュファイル書き出し
            Me_Util::write_file(DATA_ROOT . self::CACHE_DIR . $this->page_path . '.' . basename($sub_template_form->template_path), $this->output_file);
            $cache_list[$sub_template['cond']] = self::CACHE_DIR . $this->page_path . '.' . basename($sub_template_form->template_path);

            // モジュール書き出し
            foreach ($this->module_output_list as $module_name => $module_file) {
                // ファイル名のチェック
                if (!preg_match('/^[\w-.]+$/', $module_name)) {
                    // a-zA-Z-9_-. 以外の文字があれば書き出さない
                    continue;
                }
                if ($sub_template_form->char_code !== null && $sub_template_form->char_code !== Template_Logic::CHAR_CODE) {
                    $module_file = mb_convert_encoding($module_file, $sub_template_form->char_code, Template_Logic::CHAR_CODE);
                }
                // ファイル書き出し
                Me_Util::write_file(DATA_ROOT . '/module/' . $module_name, $module_file);
            }
        }
        $cache_list['_DEFAULT'] = self::CACHE_DIR . $this->page_path;

        // 公開側ページの書き出し
        // 先頭スクリプト追加
        $file_path = SITE_PATH . $this->page_path;
        // site.phpの絶対パス
        $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
        // index.phpへの相対パスを取得
        $relative_path = ME_Util::get_relative_path($file_path, $me_path);
        $output_file = "<?php include_once('" . $relative_path . "');\n"
            . '$cache_list = ' . var_export($cache_list, true) . ";\n"
            . '$cache_path = $helper->get_cache_path($cache_list);' . "\n"
            . 'include($cache_path); ?>';
        Me_Util::write_file(SITE_ROOT . $this->page_path, $output_file);
    }

    /**
     * 出力ファイル組み立て
     * @param object $node Metag_*
     * @param string $repeat_no
     * @param string $module_name
     */
    private function composition($node, $repeat_no = null, $module_name = '')
    {
        if ($node->node_type === 'module' && !empty($node->module_name)) {
            $module_name = $node->module_name;
        }

        // block要素のみ、子を呼び出す
        if ($node->is_block) {
            switch ($node->node_type) {
                case 'repeat':
                    $i = 1;
                    $count = $this->count_repeat_rows($node);
                    if (strlen($node->ref_template_id) < 1) {
                        if (is_numeric($node->offset)) {
                            $i += (integer)$node->offset;
                        }
                        if (is_numeric($node->limit)) {
                            if (is_numeric($node->offset)) {
                                $node->limit += (integer)$node->offset;
                            }
                            if ($count > $node->limit) {
                                $count = $node->limit;
                            }
                        }
                    }
                    for (; $i <= $count; $i++) {
                        foreach ($node->child_nodes as $child) {
                            $child->repeat_count = $count;
                            $this->composition($child, $i, $module_name);
                        }
                    }
                    break;
                case 'delete';
                    $exist_child_value = false;
                    $id_list = array();
                    foreach ($node->child_nodes as $child) {
                        if (strlen($child->id) > 0) {
                            $child_row = $this->get_row($child, $repeat_no, true);
                            if (!empty($child_row) && strlen($child_row['node_value']) > 0) {
                                $exist_child_value = true;
                                break;
                            }
                        }
                    }
                    if ($exist_child_value) {
                        foreach ($node->child_nodes as $child) {
                            $this->composition($child, $repeat_no, $module_name);
                        }
                    }
                    break;
                case 'display';
                    if ($node->site !== 'off') {
                        foreach ($node->child_nodes as $child) {
                            $this->composition($child, $repeat_no, $module_name);
                        }
                    }
                    break;
                default:
                    foreach ($node->child_nodes as $child) {
                        $this->composition($child, $repeat_no, $module_name);
                    }
                    break;
            }
        } else {
            // 該当ノードのデータを取り出す
            $row = $this->get_row($node, $repeat_no);

            if (!empty($module_name)) {
                $output = $node->output($row);
                $this->output_file .= $output;
                $this->module_output_list[$module_name] .= $output;
            } else {
                $this->output_file .= $node->output($row);
            }
        }

        if ($node->node_type === 'module') {
            $module_name = '';
        }
    }

    /**
     * MEタグデータ取得
     * @param object $node
     * @param string $repeat_no
     * @param boolean $check_exist_flag 存在確認フラグ
     * @return array
     */
    private function get_row($node, $repeat_no = null, $check_exist_flag = false)
    {
        $ret = null;
        // htmlノードはnullを返す。
        if ($node->node_type === 'html') {
            return $ret;
        }

        $id = (empty($node->ref_id)) ? $node->id : $node->ref_id;

        if (strlen($node->ref_page_id) > 0) {
            $data = $this->ref_page_data[$node->ref_page_id];
        } else {
            $data = $this->data;
        }

        if ($id === null || empty($node->repeat_id) || empty($repeat_no)) {
            if (isset($id) && isset($data[$id])) {
                $ret = $data[$id];
            }
        } else {
            // repeat
            $repeat_id = (empty($node->ref_repeat_id)) ? $node->repeat_id : $node->ref_repeat_id;
            if (isset($data[$repeat_id][$repeat_no][$id])) {
                $ret = $data[$repeat_id][$repeat_no][$id];
            }
        }

        // リピート領域内の場合は、空でもrepeat_noだけは返す
        if ($ret === null && $repeat_no !== null && $check_exist_flag === false) {
            $ret = array('repeat_no' => $repeat_no);
            if (strlen($node->initial_value)) {
                $ret['node_value'] = $node->initial_value;
            }
        }

        return $ret;
    }

    /**
     * repeat内データの有無確認
     * @param object $node
     * @return integer
     */
    private function count_repeat_rows($node)
    {
        if ($node->ref_page_id !== null) {
            $data = $this->ref_page_data[$node->ref_page_id];
        } else {
            $data = $this->data;
        }
        $id = (empty($node->ref_id)) ? $node->id : $node->ref_id;

        return count($data[$id]);
    }

    /**
     * idを添字にした配列に変換
     * @param array $array_list
     * @return array $assoc
     */
    private function convert_assoc_by_id($array_list)
    {
        $assoc = array();
        foreach ($array_list as $array) {
            // repeatタグの親子関係を整理
            if (!empty($array['repeat_id'])) {
                $assoc[$array['repeat_id']][$array['repeat_no']][$array['id']] = $array;
            } else {
                $assoc[$array['id']] = $array;
            }
        }
        return $assoc;
    }

    /**
     * ページ基本情報セット
     */
    private function set_basic_info()
    {
        if (!is_array($this->data)) {
            $this->data = array();
        }
        $mapping = array(
            '_page_id' => '_id',
            '_template_id' => 'template_id',
            '_path' => 'page_path',
            '_title' => 'title',
            '_keywords' => 'keywords',
            '_description' => 'description',
        );
        foreach ($mapping as $key => $prop) {
            $row = array(
                'node_value' => $this->$prop
            );
            $this->data[$key] = $row;
        }
        $this->data['_create_date'] = array('node_value' => date(Me_Config::SYSTEM_DATE_FORMAT, $this->create_date));
        $this->data['_update_date'] = array('node_value' => date(Me_Config::SYSTEM_DATE_FORMAT, $this->update_date));
        $this->data['_now'] = array('node_value' => NOW);
    }

    /**
     * $dataプロパティに配列を追加する
     * @param array $compact_data
     */
    private function set_data($compact_data)
    {
        if (!is_array($this->data)) {
            $this->data = array();
        }
        $this->data = array_merge($this->data, $compact_data);
    }

    /**
     * 空のリピート領域削除
     * @param Page_Form $form
     * @return array $data
     */
    private function delete_empty_repeat(Page_Form $form)
    {
        $data = array();
        foreach ($form->data as $key => $row) {
            $metag = $form->document->get_element_by_id($key);
            if ($metag === null || $metag->repeat_id != $row['repeat_id'] || strlen($metag->ref_id) > 0 || substr($metag->id, 0, 1) === '_') {
                // repeat_idが一致しない場合（空同士一致でも良い）はpreviewから除外
                // あるいは、参照元の場合。
                continue;
            }
            if ($metag->node_type === 'repeat') {
                $data[$key] = array();
                // inside of repeat tag
                $new_repeat_no = 1;
                foreach ($row as $repeat_no => $repeat_list) {
                    $exist = false;
                    foreach ($repeat_list as $repeat_inside_id => $repeat_inside_row) {
                        $repeat_inside_metag = $form->document->get_element_by_id($repeat_inside_id);
                        $repeat_inside_row['repeat_no'] = $new_repeat_no;
                        $repeat_inside_row['page_id'] = $form->_id;
                        // repeat内は値が無い場合は除外
                        if (strlen($repeat_inside_row['node_value']) > 0) {
                            $data[$key][$new_repeat_no][$repeat_inside_id] = $repeat_inside_row;
                            $exist = true;
                        }
                    }
                    if ($exist) {
                        $new_repeat_no++;
                    }
                }
            } else {
                $row['id'] = $key;
                $row['page_id'] = $form->_id;
                $data[$key] = $row;
            }
        }
        return $data;
    }
}